/**
 * This work was created by participants in the DataONE project, and is jointly copyrighted by participating
 * institutions in DataONE. For more information on DataONE, see our web site at http://dataone.org.
 *
 * Copyright ${year}
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 * $Id$
 */
package org.dataone.tidy;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dataone.client.MNode;
import org.dataone.cn.hazelcast.HazelcastClientFactory;
import org.dataone.configuration.Settings;
import org.dataone.service.exceptions.BaseException;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.Node;
import org.dataone.service.types.v1.NodeReference;
import org.dataone.service.types.v1.NodeType;
import org.dataone.service.types.v1.Replica;
import org.dataone.service.types.v1.Service;
import org.dataone.service.types.v1.Session;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.tidy.dao.MergeResult;
import org.dataone.tidy.dao.MergeResultRepository;
import org.dataone.tidy.dao.MergeStatus;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;

import com.hazelcast.core.IMap;
import java.util.HashMap;
import java.util.Map;
import org.dataone.service.cn.impl.v1.NodeRegistryService;
import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.ServiceFailure;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * Queries MergeResult data store for identifiers which will be evicted and re-added to the distributed system metadata
 * data structure.
 *
 * This should effectively clear any previous/old version of the system metadata object and replace it with the new
 * merged version in the distributed data strucuture. Also caused any system metadata entry listener logic to trigger
 * for the system metadata object (update the search index). Finally each MN needs to be notified of the changes to
 * system metadata - through the D1 api - systemMetadataUpdated call.
 *
 * @author sroseboo
 *
 */
public class PostMerge {

    private static Logger logger = LoggerFactory.getLogger(Main.class);
    private static final Logger successLogger = LoggerFactory.getLogger(PostMerge.class);
    private static final Logger exclusionLogger = LoggerFactory.getLogger(InclusionsFilter.class);
    private static final int PAGE_SIZE = 1000;
    private static final Long RUN_ID = Settings.getConfiguration().getLong("tidy.run.id", 999);

    private MergeResultRepository mergeResultRepository;
    private InclusionsFilter inclusionsFilter;
    private NodeRegistryService nodeRegistryService;
    
    private Map<NodeReference, Node> nodeHashMap = new HashMap<NodeReference,Node>();
    
    private PostMerge() {
    }

    public PostMerge(MergeResultRepository mergeResultRepository) {
        this.mergeResultRepository = mergeResultRepository;
        inclusionsFilter = InclusionsFilter.getInstance();
        nodeRegistryService = new NodeRegistryService();
    }

    public void postMerge() {
        PageRequest pageRequest = new PageRequest(0, PAGE_SIZE, new Sort(Sort.Direction.ASC, "id"));
        Page<MergeResult> pagedResults = mergeResultRepository.findByMergeStatusStrAndRunId(
                MergeStatus.SUCCESS.toString(), RUN_ID, pageRequest);
        processPageOfMergeResults(pagedResults.getContent());
        while (pagedResults.hasNextPage()) {
            pagedResults = mergeResultRepository.findAll(pagedResults.nextPageable());
            processPageOfMergeResults(pagedResults.getContent());
        }
    }

    private void processPageOfMergeResults(List<MergeResult> results) {
        for (MergeResult result : results) {
            SystemMetadata mergedSystemMetadata = result.getMergedSystemMetadata();
            try {
                //evictAndGetMergedPids(mergedPid);
                markMergedPidsUpdatedForMN(mergedSystemMetadata);
            } catch (PostMergeNullException ex) {
                logger.warn("MergeResult has a null systemMetadata " + result.getPid() + " because " + result.getReview());
            }
        }
    }

    private void evictAndGetMergedPids(Identifier mergedPid) {
        HazelcastClientFactory.getSystemMetadataMap().evict(mergedPid);
        HazelcastClientFactory.getSystemMetadataMap().get(mergedPid);
    }

    // Copied/modified from TransferObjectTask.auditReplicaSystemMetadata
    private void markMergedPidsUpdatedForMN(SystemMetadata mergedSystemMetadata) throws PostMergeNullException {
        try {
            if ((mergedSystemMetadata != null)
                    && inclusionsFilter.shouldInclude(mergedSystemMetadata.getIdentifier(),
                            mergedSystemMetadata.getAuthoritativeMemberNode())) {
                List<Replica> prevReplicaList = mergedSystemMetadata.getReplicaList();
                Session session = null;
                for (Replica replica : prevReplicaList) {
                    try {
                        if (! nodeHashMap.containsKey(replica.getReplicaMemberNode())) {
                            nodeHashMap.put(replica.getReplicaMemberNode(),nodeRegistryService.getNode(replica.getReplicaMemberNode()));
                        } 
                        Node node = nodeHashMap.get(replica.getReplicaMemberNode());
                        if (node.getType().equals(NodeType.MN)) {
                            boolean isTier3 = false;
                            // Find out if a tier 3 node, if not then do not callback since it is not implemented
                            for (Service service : node.getServices().getServiceList()) {
                                if (service.getName().equals("MNStorage") && service.getAvailable()) {
                                    isTier3 = true;
                                    break;
                                }
                            }
                            if (isTier3) {
                                // test again for all replicas, a Node may be down that has the 
                                // PID. if it does, we will have to create/read a special file
                                // for a retry later on
                                if(inclusionsFilter.shouldInclude(mergedSystemMetadata.getIdentifier(), node.getIdentifier())) {
                                try {
                                    String mnUrl = node.getBaseURL();
                                    MNode mnNode = new MNode(mnUrl);

                                    mnNode.systemMetadataChanged(session, mergedSystemMetadata
                                                .getIdentifier(), mergedSystemMetadata.getSerialVersion()
                                                .longValue(), mergedSystemMetadata.getDateSysMetadataModified());
                                    successLogger.info(mergedSystemMetadata.getIdentifier().getValue());
                     
                                    try {
                                        Thread.sleep(1000);
                                    } catch (InterruptedException e) {
                                        logger.error("Thread Interrupted while sleeping after system metadata changed call.");
                                    }
                                } catch (BaseException be) {
                                    logger.error(
                                            "Caught base exception when calling system metadata changed for MN: "
                                            + node.getName() + " and pid: " + mergedSystemMetadata
                                                .getIdentifier().getValue(),
                                            be);
                                }
                                } else {
                                  exclusionLogger.info(node.getIdentifier().getValue() + " " + mergedSystemMetadata.getIdentifier().getValue() );
                                }

                            }
                        }
                    } catch (ServiceFailure ex) {
                        logger.error("Caught ServiceFailure while processing PID: "
                                + mergedSystemMetadata.getIdentifier().getValue(), ex);
                    } catch (NotFound ex) {
                        logger.error("Caught NotFound while processing PID: "
                                + mergedSystemMetadata.getIdentifier().getValue(), ex);
                    }

                }
            } else {
                if (mergedSystemMetadata == null) {
                    throw new PostMergeNullException();
                }
                logger.warn("System metadata for pid: " + mergedSystemMetadata.getIdentifier().getValue()
                        + " is not included in system metadata map.");
            }
        } catch (ServiceFailure ex) {
            logger.error("Caught ServiceFailure when calling inclusionsFilter.shouldInclude  for PID: "
                    + mergedSystemMetadata.getIdentifier().getValue(), ex);
        } catch (NotFound ex) {
            logger.error("Caught NotFound when calling inclusionsFilter.shouldInclude  for PID: "
                    + mergedSystemMetadata.getIdentifier().getValue(), ex);
        }
    }
    public class PostMergeNullException extends Exception {
        
    }
}
