/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package org.dataone.tidy.merge;

import java.net.SocketTimeoutException;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.net.ssl.SSLPeerUnverifiedException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.dataone.cn.dao.exceptions.DataAccessException;
import org.dataone.configuration.Settings;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.util.DateTimeMarshaller;
import org.dataone.tidy.concurrent.TidyFutureTask;
import org.dataone.tidy.concurrent.TidyJob;
import org.dataone.tidy.concurrent.TidyJobExecutorCompletionService;
import org.dataone.tidy.concurrent.TidyJobFactory;
import org.dataone.tidy.concurrent.TidyJobThreadPoolExecutor;
import org.dataone.tidy.dao.MergeResult;
import org.dataone.tidy.dao.MergeResultRepository;
import org.dataone.tidy.dao.MergeStatus;
import org.dataone.tidy.dao.SystemMetadataTidyDao;

/**
 * At this point the Service is very tightly tied to implementations instead of interfaces making it very difficult to
 * test in isolation.
 *
 * @author waltz
 */
public class MergeExecutorService {

    private static Log logger = LogFactory.getLog(MergeExecutorService.class);

    private final int corePoolSize = Settings.getConfiguration().getInt(
            "tidy.merge.thread.pool.core.size", 10);
    private final int maximumPoolSize = Settings.getConfiguration().getInt(
            "tidy.merge.thread.pool.max.size", 10);
    private final long keepAliveSeconds = Settings.getConfiguration().getLong(
            "tidy.merge.thread.pool.keep.alive.seconds", 600L);

    private static final Long RUN_ID = Settings.getConfiguration().getLong("tidy.run.id", 999);

    private final int maxPidRetry = Settings.getConfiguration().getInt(
            "tidy.merge.executor.job.retry.max", 3);

    private final long maxFutureTimeouts = Settings.getConfiguration().getInt(
            "tidy.merge.executor.future.timeout.count.max", 6);
    private final long futureTimeoutSeconds = Settings.getConfiguration().getLong(
            "tidy.merge.executor.future.timeout.seconds", 10L);
    private final int minWorkQueueSize = Settings.getConfiguration().getInt(
            "tidy.merge.executor.queue.min.size", 100);
    private final int maxWorkQueueSize = Settings.getConfiguration().getInt(
            "tidy.merge.executor.queue.max.size", 1000);
    private CompletionService<Boolean> mergeJobCompletionService;
    private TidyJobThreadPoolExecutor mergeJobExecutor;
    private BlockingQueue workQueue = new LinkedBlockingQueue();
    private BlockingQueue<Future<Boolean>> completionQueue = new LinkedBlockingQueue<Future<Boolean>>();

    private TidyJobFactory tidyJobFactory;
    private SystemMetadataTidyDao systemMetadataDao;
    private MergeResultRepository mergeResultRepository;
    private Integer totalCompletedJobs = 0;
    private Integer totalUnrecoverableExceptions = 0;
    private Integer totalToProcess = 0;
    private ConcurrentSkipListSet<Identifier> failedPidSet = new ConcurrentSkipListSet<Identifier>();
    private Map<Identifier, Integer> retryJobPids = new HashMap<Identifier, Integer>();
    
    private MergeExecutorService() {
    }

    public MergeExecutorService(SystemMetadataTidyDao systemMetadataDao,
            MergeResultRepository mergeResultRepository, TidyJobFactory jobFactory) {

        this.systemMetadataDao = systemMetadataDao;
        this.mergeResultRepository = mergeResultRepository;
        this.tidyJobFactory = jobFactory;

        // Ignore keepAliveSeconds, If ever we get a feel for how long things are running for
        // then we can added it back in, but The executorService killing threads messes up
        // the DAO layer
//        mergeJobExecutor = new TidyJobThreadPoolExecutor(corePoolSize, maximumPoolSize,
//                keepAliveSeconds, TimeUnit.SECONDS, workQueue);
        
        /*
         * If the pool currently has more than corePoolSize threads, 
         * excess threads will be terminated if they have been idle for 
         * more than the keepAliveTime (see getKeepAliveTime(java.util.concurrent.TimeUnit)).
         * Using a value of Long.MAX_VALUE TimeUnit.NANOSECONDS effectively disables idle threads 
         * from ever terminating prior to shut down. By default, the keep-alive policy applies 
         * only when there are more than corePoolSizeThreads. 
         */
        
        mergeJobExecutor = new TidyJobThreadPoolExecutor(corePoolSize, maximumPoolSize,
                Long.MAX_VALUE, TimeUnit.NANOSECONDS, workQueue);
        //start up the task thread 
        mergeJobCompletionService = new TidyJobExecutorCompletionService<Boolean>(mergeJobExecutor,
                completionQueue);
    }

    public void executeJobs() throws DataAccessException {

        try {
            // Get the list of PIDs from the combined metacat databases
            // Assuming this a set and not a try list with multiples pid entries
            // If it is a true list with repeating entries, then it should be reduced to a set
            Set<Identifier> retryJobPidSet = new TreeSet<Identifier>();
            List<Identifier> processPidList = systemMetadataDao.listIdentifiers();
            Set<Identifier> processPidSet = new TreeSet<Identifier>(processPidList);
            Set<Identifier> priorSuccessPidSet = new TreeSet<Identifier>();
            // should subtract out the pids that have already been processed from earlier runs
            List<MergeResult> mergeResultsList = mergeResultRepository.findByMergeStatusStrAndRunId(MergeStatus.SUCCESS.toString(), RUN_ID);

            for (MergeResult mergeResult : mergeResultsList) {
                Identifier mergedPid = new Identifier();
                mergedPid.setValue(mergeResult.getPid());

                priorSuccessPidSet.add(mergedPid);

            }
            // processPidSet should only have the pids needed for the run.
            processPidSet.removeAll(priorSuccessPidSet);
            totalToProcess = processPidSet.size();
            logger.info("Total pids to process = " + totalToProcess);
            int futureTimeouts = 0;
            do {
                try {
                    // The work queue contains the jobs submitted for processing by the executor
                    // allow no less than 100 objects while the processPidSet and the
                    // retryJobPid Hash map have objects to process
                    //
                    if (workQueue.size() <= minWorkQueueSize
                            && (!processPidSet.isEmpty() || !retryJobPidSet.isEmpty())) {
                        Iterator<Identifier> identiferIterator = processPidSet.iterator();
                        int submitCount = workQueue.size();
                        Iterator<Identifier> retryIdentifierIterator = retryJobPidSet.iterator();
                        while (retryIdentifierIterator.hasNext() && (submitCount < maxWorkQueueSize)) {
                            Identifier pid = retryIdentifierIterator.next();
                            retryIdentifierIterator.remove();
                            TidyJob tidyJob = tidyJobFactory.getTidyJob(pid);
                            mergeJobCompletionService.submit(tidyJob);
                            ++submitCount;
                        }                       
                        while (identiferIterator.hasNext() && (submitCount < maxWorkQueueSize)) {

                            Identifier pid = identiferIterator.next();
                            identiferIterator.remove();

                            TidyJob tidyJob = tidyJobFactory.getTidyJob(pid);
                            mergeJobCompletionService.submit(tidyJob);

                            ++submitCount;

                        }
                    }

                    Future<Boolean> futureSuccess = mergeJobCompletionService.poll(futureTimeoutSeconds,
                            TimeUnit.SECONDS);
                    logger.debug("after submit " + completionQueue.size());
                    if (futureSuccess != null) {
                        futureTimeouts = 0;
                        Boolean reportSuccess = futureSuccess.get();

                        if (futureSuccess instanceof TidyFutureTask) {
                            TidyFutureTask tidyFuture = (TidyFutureTask) futureSuccess;
                            TidyJob tidyJob = tidyFuture.getTidyJob();
                            if (reportSuccess) {
                                ++totalCompletedJobs;
                                logger.info("Job Success: " + tidyJob.getPid().getValue());
                            } else {
                                
                                if (canHazRetry(tidyJob)) {
                                    logger.info("Job Retry: " + tidyJob.getPid().getValue());
                                    retryJobPidSet.add(tidyJob.getPid());
     
                                } else {
                                    failedPidSet.add(tidyJob.getPid());
                                    
                                    logger.error("Job Failed: " + tidyJob.getPid().getValue() + " :" + tidyJob.getFailure().getMessage() );

                                    System.err.print("[" + DateTimeMarshaller.serializeDateToUTC(new Date()) + "] Failed executeJobs: " + tidyJob.getPid().getValue() + " " );
                                    tidyJob.getFailure().printStackTrace();
                                    
                                }
                            }
                        } else {
                            // This would be a very unusual condition, don't know if it is an error
                            // state by default
                            if (reportSuccess) {
                                logger.info("Job Success but No Tidy Job returned");
                            } else {
                                logger.error("Job Failed and No Tidy Job returned");
                            }
                        }

                    } else {

                        ++futureTimeouts;
                    }
                    if (futureTimeouts > maxFutureTimeouts) {
                        logger.fatal("waited over " + (maxFutureTimeouts * futureTimeoutSeconds) + " seconds for any mergeJob to return, must be in an error state!");
                        break;
                    }
                } catch (InterruptedException e) {
                    logger.info("polling interrupted");
                } catch (ExecutionException ex) {
                    // What do we do here? if there is an exception that we can handle then we should handle and
                    // resubmit the jobs on the failed queue...
                    futureTimeouts = 0;

                    logger.error(ex);

                    // if ever we receive a SocketTimeoutException,
                    // then go threw all the failures contained in 
                    // the mergeJobExecutor
                    // and review all jobs to see if which ones
                    // have failed from a SocketTimeoutException
                    // so as to resubmit them in the hope that they will now work
                    if (hasTransientThrowable(ex)) {

                        TidyJob retryJob = null;
                        while ((retryJob = mergeJobExecutor.getFailure()) != null) {

                            logger.info("ExecutionException retryJob has " + retryJob.getPid().getValue());
                            if (!canHazRetry(retryJob)) {
                                failedPidSet.add(retryJob.getPid());
                                System.err.print("[" + DateTimeMarshaller.serializeDateToUTC(new Date()) + "]  no retry executeJobs: " + ex.getClass().getCanonicalName() + " " );
                                ex.printStackTrace();
                            } else {
                                retryJobPidSet.add(retryJob.getPid());
                            }      
                        }

                    } else {
                        ++totalUnrecoverableExceptions;
                        System.err.print("[" + DateTimeMarshaller.serializeDateToUTC(new Date()) + "] unrecoverable executeJob: " + ex.getClass().getCanonicalName() + " " );
                        ex.printStackTrace();
                    }
                }
                if (mergeJobExecutor.getActiveCount() > maximumPoolSize) {
                    logger.error("Job Executor has more jobs that can be handled! " + mergeJobExecutor.getActiveCount());
                }
                logger.debug("TotalToProcess = " + Integer.toString(totalToProcess) + " totalCompleted " + Integer.toString(totalCompletedJobs) + " totalFailed " + Integer.toString(failedPidSet.size()) + " total Unrecoverable "  +  Integer.toString(totalUnrecoverableExceptions));
            } while ((mergeJobExecutor.getActiveCount() != 0) || (totalToProcess > (totalCompletedJobs + totalUnrecoverableExceptions + failedPidSet.size())));


            
        //
            // given the nature of concurrency, there might be a failure that was not found in the iterator
            // while looping threw the mergeJobExecutor.getFailures after an exception was thrown.
            //
            // may need to review mergeJobExecutor.getFailures to make certain all have been processed
            // and the same with the retryJobPids.
            // report on total number of failures...
        } catch (DataAccessException dae) {
            // anyway to recover from this?
            dae.printStackTrace();
        } catch (Exception ex) {
            // probably a null pointer exception from somewhere
            ex.printStackTrace();
        } finally {

            logger.info("Number of Jobs Successfuly Completed " + totalCompletedJobs);
            logger.info("Number of Jobs on Failed " + failedPidSet.size());
            logger.info("Number of Jobs with Unrecoverable Exceptions " + totalUnrecoverableExceptions);  
            mergeJobExecutor.shutdownNow();
        }

    }

    public Integer getTotalCompletedJobs() {
        return totalCompletedJobs;
    }

    public Integer getTotalFailedJobs() {
        return totalUnrecoverableExceptions + failedPidSet.size();
    }

    public Integer getTotalToProcess() {
        return totalToProcess;
    }

    private Boolean canHazRetry(TidyJob retryJob) {
        Boolean retry = false;
        if (!(failedPidSet.contains(retryJob.getPid()))  && hasTransientThrowable(retryJob.getFailure())) {

            if (!retryJobPids.containsKey(retryJob.getPid())) {
                retryJobPids.put(retryJob.getPid(), 0);
                retry = true;
                logger.info("retry " + retryJob.getPid().getValue() + " because of SocketTimeoutException");
            } else {

                Integer failedCount = retryJobPids.get(retryJob
                        .getPid());
                ++failedCount;
                logger.info("retry " + retryJob.getPid().getValue() + " with count " + failedCount);
                // only retry a maximum # of times before complete and utter failure (3 strikes and you're out)
                if (failedCount >= maxPidRetry) {
                    logger.warn("fail count exceeded removing job from retry, adding to failedPidSet: " + retryJob.getPid().getValue());
                    retryJobPids.remove(retryJob.getPid());

                } else {
                    retryJobPids.put(retryJob.getPid(), failedCount);
                    retry = true;
                }
            }
        }
        return retry;
    }

    private Boolean hasTransientThrowable(Throwable t) {

        // regarding causes in Exceptions:
        // though initcause can not be set with "this",
        // there is a way to create an infinite loop in causes
        //
        // http://www.coderanch.com/t/569787/java/java/exception-getCause
        // Exception yin = new Exception();  
        // Exception yang = new Exception();  
        // yin.initCause(yang);  
        // yang.initCause(yin);
        //
        Boolean isTransient = false;

        // TreeSet throws null pointer exception if natural ordering is used (as by the default)
        // and a null is added to it.
        Set<Throwable> causeList = new HashSet<Throwable>();
        if (t != null) 
            logger.debug(t.getMessage());
        // first one will always work or throw an exception
        Boolean addedCause = causeList.add(t);
        do {
           if ( (t instanceof SocketTimeoutException) || (t instanceof org.apache.http.conn.ConnectTimeoutException) || (t instanceof org.apache.commons.httpclient.ConnectTimeoutException) || (t instanceof SSLPeerUnverifiedException) ) {
                isTransient = true;
            }
            t = t.getCause();
            if (t != null) 
                logger.debug(t.getMessage());
            addedCause = causeList.add(t);

        } while ((t != null) && !(isTransient) && addedCause);

        return isTransient;
    }
}
